import logging
import os
import pickle
import re
import time
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import openai
import tiktoken
from tqdm import tqdm


@dataclass
class TCC:
    """OpenAI Chat Completion.

    Parameters
    ----------
    input_dir : str
        Path to the input directory including text files.
    input_suffix: str
        Comma separated suffixes of input files, by default "txt,md,markdown"
    output_file : str
        Path to the output pickle file.
    key : str
        OpenAI API key. If not given, try to get from environment variable: OPENAI_API_KEY.
    character_encoding: str
        Character encoding for input file, by default "utf-8"
    chat_model: str
        Chat model name, by default "gpt-3.5-turbo"
    encoding : str
        Encoding name for tiktoken , by default "cl100k_base"
    embedding : str
        Embedding model name, by default "text-embedding-ada-002"
    remain_url: bool
        Set to True to keep URL in the text, by default False (replace URL as "URL"
    keep_spaces: bool
        Set to True to keep spaces in the text, by default False (replace to single space)
    block_size : int
        Block size for embedding, by default 500
    embed_max_size : int
        Max size for embedding, by default 8150
    max_prompt_size : int
        Max size for prompt, by default 4096
    return_size : int
        Return size, by default 250
    prompt : str
        Prompt template.
    bare_prompt: str
        Prompt template without index.
    question : str
        Question template.
    no_index: bool
        Set to True to ask the question directly
    verbose: bool
        Set to True to show detailed log
    """

    input_dir: str = ""
    input_suffix: str = "txt,md,markdown"
    output_file: str = "index.pickle"
    key: str = ""
    character_encoding: str = "utf-8"
    chat_model: str = "gpt-3.5-turbo"
    encoding: str = "cl100k_base"
    embedding: str = "text-embedding-ada-002"
    remain_url: bool = False
    keep_spaces: bool = False
    block_size: int = 500
    embed_max_size: int = 8150  # actual limit is 8191
    max_prompt_size: int = 4096
    return_size: int = 250
    prompt: str = """Read the following text and answer the question. Your reply should be shorter than RETURN_SIZE characters.

## Text
{text}

## Question
{question}"""
    bare_prompt: str = """Answer the question. Your reply should be shorter than RETURN_SIZE characters.

## Question
{question}"""
    question: str = "What is the most important question?"
    no_index: bool = False
    verbose: bool = False

    def __post_init__(self) -> None:
        self.log = logging.getLogger(__name__)
        if self.verbose:
            if self.log.parent and self.log.parent.name != "root":
                self.log.parent.setLevel(logging.DEBUG)
            else:
                self.log.setLevel(logging.DEBUG)

        self.enc = tiktoken.get_encoding(self.encoding)
        self.prompt = self.prompt.replace(
            "RETURN_SIZE", str(self.return_size)
        ).strip()
        self.bare_prompt = self.bare_prompt.replace(
            "RETURN_SIZE", str(self.return_size)
        ).strip()
        self.question = self.question.strip()
        if not self.output_file.endswith(".pickle"):
            self.output_file = self.output_file + ".pickle"
        try:
            self.cache = pickle.load(open(self.output_file, "rb"))
        except FileNotFoundError:
            self.cache = {}

    def set_key(self, key: str = "") -> bool:
        if key:
            self.key = key
        if not self.key:
            self.key = os.environ.get("OPENAI_API_KEY", "")
        if not self.key:
            self.log.error(
                "Set OPEN_AI_API_KEY environment variable or use -k option"
            )
            return False
        openai.api_key = self.key
        return True

    def get_size(self, text: str) -> int:
        return len(self.enc.encode(text))

    def embed(self, text: str) -> list[float]:
        text = text.replace("\n", " ")
        tokens = self.enc.encode(text)
        if len(tokens) > self.embed_max_size:
            text = self.enc.decode(tokens[: self.embed_max_size])

        while True:
            try:
                res = openai.Embedding.create(
                    input=[text], model=self.embedding
                )
                time.sleep(1)
            except Exception as e:
                self.log.debug(e)
                time.sleep(1)
                continue
            break

        return res["data"][0]["embedding"]

    def get_files(self) -> list[Path]:
        path = Path(self.input_dir)
        files = []
        for suffix in self.input_suffix.split(","):
            files += list(path.glob(f"**/*.{suffix}"))
        if not files:
            self.log.error(
                f"No {', '.join(['*.' + x for x in self.input_suffix.split(',')])} found in {self.input_dir}."
            )
        return sorted(files, key=lambda x: x.name)

    def get_or_make(self, body: str, title: str) -> tuple[list[float], str]:
        if body not in self.cache:
            self.cache[body] = (self.embed(body), title)
            pickle.dump(self.cache, open(self.output_file, "wb"))
        return self.cache[body]

    def update_from_text(self) -> int:
        if not self.set_key():
            return 1
        if not self.input_dir:
            self.log.error("Set input directory")
            return 1
        files = self.get_files()
        if not files:
            self.log.error(
                f"No {', '.join(['*.' + x for x in self.input_suffix.split(',')])} found in {self.input_dir}."
            )
            return 1
        for f in tqdm(sorted(files)):
            buf = []
            title = f.name
            with open(f) as fp:
                for line in fp.readlines():
                    line = line.strip()
                    if not self.remain_url:
                        line = re.sub(r"https?://[^\s]+", "URL", line)
                    if not self.keep_spaces:
                        line = re.sub(r"[\s]+", " ", line)
                    buf.append(line)
                    body = " ".join(buf)
                    if self.get_size(body) > self.block_size:
                        self.get_or_make(body, title)
                        buf = buf[len(buf) // 2 :]
            body = " ".join(buf).strip()
            if body:
                self.get_or_make(body, title)
        return 0

    def get_sorted(self, query: str) -> list[tuple[float, str, str]]:
        q = np.array(self.embed(query))
        buf = []
        for body, (v, title) in tqdm(
            self.cache.items(), disable=not self.verbose
        ):
            buf.append((q.dot(v), body, title))
        buf.sort(reverse=True)
        return buf

    def ask(self) -> int:
        if not self.set_key():
            return 1
        prompt = self.bare_prompt if self.no_index else self.prompt
        prompt_size = self.get_size(prompt)
        rest = self.max_prompt_size - self.return_size - prompt_size
        if not self.question:
            self.log.error("Ask question!")
            return 1
        input_size = self.get_size(self.question)
        if rest < input_size:
            self.log.error("too large input!")
            return 1
        rest -= input_size

        if not self.no_index:
            samples = self.get_sorted(self.question)

            to_use = []
            used_title = []
            for _sim, body, title in samples:
                if title in used_title:
                    continue
                size = self.get_size(body)
                if rest < size:
                    break
                to_use.append(body)
                used_title.append(title)
                rest -= size
            text = "\n\n".join(to_use)
            prompt = prompt.format(question=self.question, text=text)
        else:
            prompt = prompt.format(question=self.question)

        self.log.debug("\nPROMPT:")
        self.log.debug(prompt)

        self.log.debug("\nTHINKING...")
        response = openai.ChatCompletion.create(
            model=self.chat_model,
            messages=[{"role": "user", "content": prompt}],
            max_tokens=self.return_size,
            temperature=0.0,
        )

        self.log.debug("\nRESPONSE:")
        self.log.debug(response)
        content = response["choices"][0]["message"]["content"]

        self.log.debug("\nANSWER:")
        self.log.info(f">>>> {self.question}")
        self.log.info(f"> {content}")
        if not self.no_index:
            self.log.info(f"\nref. {', '.join(used_title)}")
        return 0



import argparse
import logging
import sys

from .__version__ import __version__
from .tcc import TCC


def main() -> int:
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    log = logging.getLogger(__name__.split(".")[0])
    log.setLevel(logging.INFO)

    tcc = TCC()

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "command",
        help="Command (index or ask)",
        type=str,
        nargs="?",
    )
    arg_parser.add_argument(
        "-i",
        "--input_dir",
        help="Input directory",
        type=str,
    )
    arg_parser.add_argument(
        "-s",
        "--input_suffix",
        help=f'Comma separated suffixes of input files, "{tcc.input_suffix}"',
        type=str,
    )
    arg_parser.add_argument(
        "-o",
        "--output_file",
        help=f'Output file (pickle), default: "{tcc.output_file}"',
        type=str,
    )
    arg_parser.add_argument(
        "-k",
        "--key",
        help="OpenAI API key. If not given, try to get from environment variable: OPENAI_API_KEY.",
        type=str,
    )
    arg_parser.add_argument(
        "-c",
        "--character_encoding",
        help=f'Character encoding for input file, default: "{tcc.character_encoding}"',
        type=str,
    )
    arg_parser.add_argument(
        "--chat_model",
        help=f'Chat model name, default: "{tcc.chat_model}"',
        type=str,
    )
    arg_parser.add_argument(
        "--encoding",
        help=f'Encoding name for tiktoken, default: "{tcc.encoding}"',
        type=str,
    )
    arg_parser.add_argument(
        "--embedding",
        help=f'Embedding model name, default: "{tcc.embedding}"',
        type=str,
    )
    arg_parser.add_argument(
        "--remain_url",
        action="store_true",
        help="Keep URL in the text",
    )
    arg_parser.add_argument(
        "--keep_spaces",
        action="store_true",
        help="Keep spaces in the text",
    )
    arg_parser.add_argument(
        "--block_size",
        help=f"Block size for embedding, default: {tcc.block_size}",
        type=int,
    )
    arg_parser.add_argument(
        "--embed_max_size",
        help=f"Max size for embedding, default: {tcc.embed_max_size}",
        type=int,
    )
    arg_parser.add_argument(
        "--max_prompt_size",
        help=f"Max size for prompt, default: {tcc.max_prompt_size}",
        type=int,
    )
    arg_parser.add_argument(
        "--return_size",
        help=f"Return size, default: {tcc.return_size}",
        type=int,
    )
    arg_parser.add_argument(
        "--prompt",
        help=f'Prompt template, default: "{tcc.prompt}"',
        type=str,
    )
    arg_parser.add_argument(
        "--bare_prompt",
        help=f'Prompt template without index, default: "{tcc.prompt}"',
        type=str,
    )
    arg_parser.add_argument(
        "-q",
        "--question",
        help=f'Question words for ask, default: "{tcc.question}"',
        type=str,
    )
    arg_parser.add_argument(
        "-n",
        "--no_index",
        action="store_true",
        help="Ask the question directly",
    )
    arg_parser.add_argument(
        "-V",
        "--verbose",
        action="store_true",
        help="Show detailed log",
    )
    arg_parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="Show version",
    )

    args = arg_parser.parse_args()
    params = {}
    command = None
    for x in args.__dict__:
        if x == "version":
            if args.__dict__[x]:
                log.info(f"tcc {__version__}")
                return 0
        elif x == "command":
            command = args.__dict__[x]
        elif args.__dict__[x] is not None:
            params[x] = args.__dict__[x]

    if not command:
        log.error("No command specified")
        return 1

    tcc = TCC(**params)

    if args.command not in ["index", "ask"]:
        log.error(f"Invalid command: {args.command}. Must be index or ask")
        return 1
    elif args.command == "index":
        return tcc.update_from_text()
    elif args.command == "ask":
        return tcc.ask()

    return 0


if __name__ == "__main__":
    sys.exit(main())
